/***

Emu48 DDE Client to Server transfer

Copyright (C) 2000-2006 Christoph Gieelink

This is free software including the sources and can be used freely.
You're allowed to modify or to use this file complete or in parts
in your own applications without any author notify.


Technical data:

Servername:      Emu48
Topicname:       Stack
Item:            1 (stack level)
Clipboardformat: "CF_HPOBJ" (user defined)

The DDE commands CONNECT, POKE and REQUEST are supported.

The structure of the clipboard format "CF_HPOBJ":

+--------+------------------------------------+
| 4 Byte |             HP object              |
+--------+------------------------------------+
    \        \
     \        +--- normal HP object
      +----------- length of object (LSB first)

***/

#define STRICT
#include <windows.h>
#include <ddeml.h>

#define CF_HPOBJ  "CF_HPOBJ"				// clipboard format for DDE

static HWND   hWndEmu48 = NULL;				// handle of Emu48 window
static LPTSTR szEmuWnd = TEXT("CEmu48");	// class name of Emu48 window

static BOOL CALLBACK EnumThreadWndProc(HWND hWnd,LPARAM lParam)
{
	BOOL  bFound;
	TCHAR szClass[8];

	// Emu48 window found?
	GetClassName(hWnd,szClass,sizeof(szClass)/sizeof(*szClass));
	if (bFound = (lstrcmp(szClass,szEmuWnd) == 0))
		hWndEmu48 = hWnd;

	return !bFound;
	UNREFERENCED_PARAMETER(lParam);
}

BOOL DdeSend(LPCTSTR lpszService,UINT nStkLvl,LPBYTE lpData) // FALSE = transfer OK, TRUE = transfer failed
{
	DWORD  idDdeInst = 0;					// DDE server id
	HSZ    hszService, hszTopic, hszItem;	// DDE server variables
	HCONV  hConv;							// DDE connection handle
	UINT   uCF_HpObj;						// DDE clipboard format
	TCHAR  szItem[8];						// DDE item string

	BOOL   bErr = TRUE;						// error

	// get number of bytes to send
	DWORD dwSize = sizeof(DWORD) + *(DWORD *)lpData;

	if (DdeInitialize(&idDdeInst,(PFNCALLBACK) NULL,
					  APPCMD_CLIENTONLY | CBF_SKIP_ALLNOTIFICATIONS,0) != DMLERR_NO_ERROR)
		return TRUE;						// init fail

	wsprintf(szItem,TEXT("%u"),nStkLvl);	// stack level as DDE item

	// init clipboard format
	uCF_HpObj = RegisterClipboardFormat(TEXT(CF_HPOBJ));

	// init DDE client handles
	hszService = DdeCreateStringHandle(idDdeInst,lpszService,0);
	hszTopic   = DdeCreateStringHandle(idDdeInst,TEXT("Stack"),0);
	hszItem    = DdeCreateStringHandle(idDdeInst,szItem,0);

	// connect to Emu48 DDE server
	if ((hConv = DdeConnect(idDdeInst,hszService,hszTopic,NULL)) != NULL)
	{
		// transfer data to server
		if (DdeClientTransaction(lpData,dwSize,hConv,hszItem,uCF_HpObj,XTYP_POKE,5000,NULL) != NULL)
		{
			CONVINFO ConvInfo;

			// get window handle of communication partner
			ConvInfo.cb = sizeof(ConvInfo);
			DdeQueryConvInfo(hConv,QID_SYNC,&ConvInfo);

			// get CEmu48 window handle from partner
			EnumThreadWindows(GetWindowThreadProcessId(ConvInfo.hwndPartner,NULL),&EnumThreadWndProc,0);

			bErr = FALSE;					// data tranfer successful
		}
		DdeDisconnect(hConv);				// disconnect DDE server
	}

	// clean up DDE client
	DdeFreeStringHandle(idDdeInst,hszService);
	DdeFreeStringHandle(idDdeInst,hszTopic);
	DdeFreeStringHandle(idDdeInst,hszItem);
	DdeUninitialize(idDdeInst);
	return bErr;
}


#include <stdio.h>

INT main(INT argc, CHAR *argv[])
{
	HANDLE hFile;
	LPBYTE lpBuf;
	DWORD  dwFileSizeLow,dwFileSizeHigh;
	CHAR   *lpszService,*lpszFileName;
	BOOL   bErr;

	if (!(argc == 2 || (argc == 3 && *argv[1] == '/')))
	{
		printf(TEXT("Syntax: DDE48 [/Servicename] filename\n"));
		return 1;
	}

	if (argc == 3 && *argv[1] == '/')
	{
		static CHAR szClass[256] = "C";

		lpszService  = argv[1]+1;			// name after "/"
		lpszFileName = argv[2];				// filename

		// class name
		strncpy(&szClass[1],lpszService,sizeof(szClass)/sizeof(szClass[0])-1);
		szEmuWnd = szClass;
	}
	else
	{
		lpszService  = "Emu48";
		lpszFileName = argv[1];
	}

	// copy file to buffer
	hFile = CreateFile(lpszFileName,GENERIC_READ,0,NULL,OPEN_EXISTING,FILE_FLAG_SEQUENTIAL_SCAN,NULL);
	if(hFile == INVALID_HANDLE_VALUE)
	{
		printf("Error: File %s not found!\n",lpszFileName);
		return 1;
	}
	dwFileSizeLow = GetFileSize(hFile, &dwFileSizeHigh);
	if (dwFileSizeHigh != 0)
	{  // file is too large.
		printf("Error: File %s too large!\n",lpszFileName);
		CloseHandle(hFile);
		return 2;
	}
	lpBuf = (LPBYTE)LocalAlloc(LMEM_FIXED,dwFileSizeLow+4);
	if (lpBuf == NULL)
	{
		printf("Error: Allocation problem!\n");
		CloseHandle(hFile);
		return 2;
	}

	// make "CF_HPOBJ" data object
	*(DWORD *)lpBuf = dwFileSizeLow;		// length of object (LSB first)
	ReadFile(hFile,lpBuf+sizeof(DWORD),dwFileSizeLow,&dwFileSizeHigh,NULL);

	bErr = DdeSend(lpszService,1,lpBuf);	// transfer file to stack level 1
	printf("Transfer: %s\n",bErr ? "failed" : "OK");

	CloseHandle(hFile);
	LocalFree(lpBuf);

	if (hWndEmu48 != NULL)					// activate Emu48 window
	{
		ShowWindow(hWndEmu48,SW_SHOWNORMAL);
		SetForegroundWindow(hWndEmu48);
	}

	return 0;
}
